

class ChatBotException(Exception):
    pass


class WrongCreds(ChatBotException):
    pass


class HookException(ChatBotException):
    pass


class APIException(ChatBotException):
    pass


class BotNotRunning(ChatBotException):
    pass



from .chatbot import ChatBot
from . import reddit_auth as RedditAuthentication
from ._utils.frame_model import FrameType
from ._api.models import CustomType, MemberState
from ._api.iconkeys import Snoo, Reaction
from ._utils._exceptions import *



from enum import Enum


class Snoo(str, Enum):
    SNOO_ANGRY = 'snoo_angry'
    SNOO_FEELSBADMAN = 'snoo_feelsbadman'
    UPVOTE = 'upvote'
    SNOO_HEARTEYES = 'snoo_hearteyes'
    SNOO_FEELSGOODMAN = 'snoo_feelsgoodman'
    KAPPA = 'kappa'
    SNOO_TONGUE = 'snoo_tongue'
    CAT_BLEP = 'cat_blep'
    SNOO = 'snoo'
    SNOO_CRY = 'snoo_cry'
    SNOO_WINK = 'snoo_wink'
    SNOO_SAD = 'snoo_sad'
    REDDITGOLD = 'redditgold'
    SNOO_THOUGHTFUL = 'snoo_thoughtful'
    CAKE = 'cake'
    SNOO_SURPRISED = 'snoo_surprised'
    UPVOTE_PRIDE = 'upvote_pride'
    SNOO_JOY = 'snoo_joy'
    GIVE_UPVOTE_PRIDE = 'give_upvote_pride'
    SNOO_PUTBACK = 'snoo_putback'
    SNOO_FACEPALM = 'snoo_facepalm'
    SNOO_SIMPLE_SMILE = 'snoo_simple_smile'
    SNOO_SCREAM = 'snoo_scream'
    SNOO_SMILE = 'snoo_smile'
    ILLUMINATI = 'illuminati'
    PUPPER = 'pupper'
    SNOO_SHRUG = 'snoo_shrug'
    DOWNVOTE = 'downvote'
    SLOTH = 'sloth'
    SNOO_HUG = 'snoo_hug'
    SNOO_TROLLFACE = 'snoo_trollface'
    HEART_EYES_PRIDE = 'heart_eyes_pride'
    DOGE = 'doge'
    SNOO_DEALWITHIT = 'snoo_dealwithit'
    SNOO_DISAPPROVAL = 'snoo_disapproval'
    PARTYPARROT = 'partyparrot'
    SNOO_TABLEFLIP = 'snoo_tableflip'
    KARMA = 'karma'
    ORLY = 'orly'
    HAMSTER = 'hamster'


class Reaction(str, Enum):
    REACT1 = 'mi2jolzfa7081.gif'
    REACT2 = '2o3aooqfa7081.gif'
    REACT3 = 'k7ry7t1ga7081.gif'
    REACT4 = 'jvuspmbga7081.gif'
    REACT5 = 'g6akcitga7081.gif'
    REACT6 = '3i2arwzga7081.gif'
    REACT7 = 'cdnhvqfga7081.gif'
    REACT8 = 't5bqwxdyt7081.gif'
    REACT9 = 'tspuf53ga7081.gif'
    REACT10 = 'ax7wu47ga7081.gif'
    REACT11 = 't2r5xc9ga7081.gif'
    REACT12 = '0ku6twega7081.gif'
    REACT13 = 'b5s6cohga7081.gif'
    REACT14 = 'pleyoikga7081.gif'
    REACT15 = 'zn7iubvfa7081.gif'
    REACT16 = '19b5q4vga7081.gif'
    REACT17 = '8r21ukpfa7081.gif'
    REACT18 = 'm7uy86lga7081.gif'
    REACT19 = 'wbrgz1nga7081.gif'
    REACT20 = 'vq7naqwfa7081.gif'
    REACT21 = 'qzl5vyxfa7081.gif'
    REACT22 = 'sjs1a2fyt7081.gif'
    REACT23 = '9ut7iedga7081.gif'
    REACT24 = '3s0glewga7081.gif'
    REACT25 = '00brcfjga7081.gif'
    REACT26 = 'dfxygs4ga7081.gif'
    REACT27 = '1p1jgpgga7081.gif'
    REACT28 = 't1djdguga7081.gif'
    REACT29 = '79opsq0ha7081.gif'
    REACT30 = 'foyijyyga7081.gif'
    REACT31 = 'uy83aa8ga7081.gif'
    REACT32 = 'mag7v6tfa7081.gif'
    REACT33 = 'xjls1pqga7081.gif'
    REACT34 = 'ksz4fmaga7081.gif'
    REACT35 = 'evwks24ga7081.gif'
    REACT36 = 'd45pfmsga7081.gif'
    REACT37 = 'iuqmp7ufa7081.gif'
    REACT38 = '8zx5ixoga7081.gif'
    REACT39 = 'ytv3x0sfa7081.png'
    REACT40 = '4zhlw4iyt7081.gif'
    REACT41 = 'av9z8iiga7081.gif'
    REACT42 = '7rc03trga7081.gif'
    REACT43 = 'mp9zclcga7081.gif'
    REACT44 = 'fsg2a1oga7081.gif'
    REACT45 = '7url39xga7081.gif'
    REACT46 = 'd2kn6yxga7081.gif'
    REACT47 = '8kw138jyt7081.gif'
    REACT48 = 'ul2w17ega7081.gif'



import requests
from .._utils.consts import *
from .._utils._exceptions import APIException, BotNotRunning
import json
from .models import Channel, Message, Members, BannedUsers
from typing import Callable, Optional, Union, List
import logging
from .iconkeys import Reaction


def _get_user_id(username):
    response = requests.get(f"{WWW_REDDIT}/user/{username}/about.json", headers={'user-agent': WEB_USERAGENT}).json()
    u_id = response.get('data', {}).get('id')
    if u_id is None:
        return None
    else:
        return f't2_{u_id}'


class Tools:
    def __init__(self, reddit_auth, session_key_getter: Callable[[], str], is_running_getter: Callable[[], bool]):
        self.__reddit_auth = reddit_auth
        self.__req_sesh = requests.Session()
        self.__session_key_getter: Optional[Callable[[], str]] = session_key_getter
        self.__is_running_getter: Callable[[], bool] = is_running_getter

    def __handled_req(self, method: str, uri: str, chatmedia: bool, **kwargs) -> requests.Response:
        while self.__is_running_getter():
            if chatmedia:
                headers = {
                    'User-Agent': USER_AGENT,
                    'SendBird': f'Android,31,3.0.168,{SB_ai}',
                    'SB-User-Agent': SB_User_Agent,
                    'Session-Key': self.__session_key_getter()
                }
            else:
                headers = {
                    'User-Agent': MOBILE_USERAGENT,
                    'Authorization': f'Bearer {self.__reddit_auth.api_token}',
                    'Content-Type': 'application/json',
                }
            response = self.__req_sesh.request(method, uri, headers=headers, **kwargs)
            if response.status_code == 401 and self.__reddit_auth.is_reauthable:
                new_access_token = self.__reddit_auth.refresh_api_token()
                headers.update({'Authorization': f'Bearer {new_access_token}'})
                continue
            elif response.status_code != 200:
                raise APIException(logging.error(response.text))
            else:
                return response
        raise BotNotRunning("Cannot do that without running the bot first")

    def send_reaction(self, reaction_icon_key: Reaction, msg_id: Union[str, int], channel_url: str) -> None:
        data = json.dumps({
            "id": "7628b2213978",
            "variables": {
                "channelSendbirdId": channel_url,
                "messageSendbirdId": str(msg_id),
                "reactionIconKey": reaction_icon_key.value,
                "type": "ADD"
            }
        })
        self.__handled_req(method='POST', uri=GQL_REDDIT, chatmedia=False, data=data)

    def delete_reaction(self, reaction_icon_key: Reaction, msg_id: Union[str, int], channel_url: str) -> None:
        data = json.dumps({
            "id": "7628b2213978",
            "variables": {
                "channelSendbirdId": channel_url,
                "messageSendbirdId": msg_id,
                "reactionIconKey": reaction_icon_key.value,
                "type": "DELETE"
            }
        })
        self.__handled_req(method='POST', uri=GQL_REDDIT, chatmedia=False, data=data)

    def rename_channel(self, name: str, channel_url: str) -> Channel:
        data = json.dumps({'name': name})
        response = self.__handled_req(method='PUT', uri=f"{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}",
                                      chatmedia=True,
                                      data=data)
        return Channel(response.json())

    def delete_message(self, channel_url: str, msg_id: int) -> None:
        self.__handled_req(method='DELETE',
                           uri=f"{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/messages/{msg_id}",
                           chatmedia=True)

    def kick_user(self, channel_url: str, user_id: str, duration: int) -> None:
        data = json.dumps({
            'channel_url': channel_url,
            'user_id': user_id,
            'duration': duration
        })
        self.__handled_req(method='POST', uri=f'{S_REDDIT}/api/v1/channel/kick/user',
                           chatmedia=False,
                           data=data)

    def invite_user_to_channel(self, channel_url: str, nicknames: Union[str, List[str]]) -> None:
        if isinstance(nicknames, str):
            nicknames = [nicknames]
        users = []
        for nickname in nicknames:
            users.append({'user_id': _get_user_id(nickname), 'nickname': nickname})
        data = json.dumps({'users': users})
        self.__handled_req(method='POST', uri=f'{S_REDDIT}/api/v1/sendbird/group_channels/{channel_url}/invite',
                           chatmedia=False,
                           data=data)

    def accept_chat_invite(self, channel_url):
        data = json.dumps({
            'user_id': self.__reddit_auth.user_id
        })
        response = self.__handled_req(method='PUT', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/accept',
                                      chatmedia=True, data=data)
        return Channel(response.json())

    def get_channels(self, limit: int = 100, order: str = "latest_last_message", show_member: bool = True,
                     show_read_receipt: bool = True, show_empty: bool = True, member_state_filter: str = "joined_only",
                     super_mode: str = "all", public_mode: str = "all", unread_filter: str = "all",
                     hidden_mode: str = "unhidden_only", show_frozen: bool = True,
                     # custom_types: str = 'direct,group'
                     ) -> List[Channel]:
        params = {
            'limit': limit,
            'order': order,
            'show_member': show_member,
            'show_read_receipt': show_read_receipt,
            'show_delivery_receipt': 'true',
            'show_empty': show_empty,
            'member_state_filter': member_state_filter,
            'super_mode': super_mode,
            'public_mode': public_mode,
            'unread_filter': unread_filter,
            'hidden_mode': hidden_mode,
            'show_frozen': show_frozen,
        }
        response = self.__handled_req(method='GET',
                                      uri=f'{SB_PROXY_CHATMEDIA}/v3/users/{self.__reddit_auth.user_id}/my_group_channels',
                                      chatmedia=True, params=params)
        return [Channel(channel) for channel in response.json()['channels']]

    def get_members(self, channel_url: str, next_token: str = None, limit: int = 20,
                    order: str = "member_nickname_alphabetical", member_state_filter: str = "all",
                    nickname_startswith: str = '') -> Members:
        params = {
            'token': next_token,
            'limit': limit,
            'order': order,
            'muted_member_filter': 'all',
            'member_state_filter': member_state_filter,
            'show_member_is_muted': 'true',
            'show_read_receipt': 'true',
            'show_delivery_receipt': 'true',
            'nickname_startswith': nickname_startswith
        }
        response = self.__handled_req(method='GET', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/members',
                                      chatmedia=True, params=params)
        return Members(response.json())

    def get_banned_members(self, channel_url, limit: int = 100) -> BannedUsers:
        params = {'limit': limit}
        response = self.__handled_req(method='GET', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/ban',
                                      chatmedia=True, params=params)
        return BannedUsers(response.json())

    def leave_chat(self, channel_url: str) -> None:
        data = json.dumps({
            'user_id': self.__reddit_auth.user_id
        })
        self.__handled_req(method='PUT', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/leave',
                           chatmedia=True, data=data)

    def create_channel(self, nicknames: List[str], group_name: str) -> Channel:
        users = []
        for nickname in nicknames:
            users.append({'user_id': _get_user_id(nickname), 'nickname': nickname})
        data = json.dumps({
            'users': users,
            'name': group_name
        })
        response = self.__handled_req(method='POST', uri=f'{S_REDDIT}/api/v1/sendbird/group_channels',
                                      chatmedia=False,
                                      data=data)
        return Channel(response.json())

    def hide_chat(self, channel_url: str, hide_previous_messages: bool = False, allow_auto_unhide: bool = True) -> None:
        data = json.dumps({
            'user_id': self.__reddit_auth.user_id,
            'hide_previous_messages': hide_previous_messages,
            'allow_auto_unhide': allow_auto_unhide
        })
        self.__handled_req(method='PUT', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/hide',
                           chatmedia=True, data=data)

    def unhide_chat(self, channel_url: str) -> None:
        self.__handled_req(method='DELETE', uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/hide',
                           chatmedia=True)

    def get_older_messages(self, channel_url: str, message_ts: Union[int, str] = 9007199254740991,
                           custom_types: str = '*', prev_limit: int = 40, next_limit: int = 0,
                           reverse: bool = True) -> List[Message]:
        params = {
            'is_sdk': 'true',
            'prev_limit': prev_limit,
            'next_limit': next_limit,
            'include': 'false',
            'reverse': reverse,
            'message_ts': message_ts,
            'custom_types': custom_types,
            'with_sorted_meta_array': 'false',
            'include_reactions': 'false',
            'include_thread_info': 'false',
            'include_replies': 'false',
            'include_parent_message_text': 'false'
        }
        response = self.__handled_req(method='GET',
                                      uri=f'{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/messages',
                                      chatmedia=True, params=params)
        return [Message(msg) for msg in response.json()['messages']]

    def mute_user(self, channel_url: str, user_id: str, duration: int, description: str) -> None:
        data = json.dumps({
            'user_id': user_id,
            'seconds': duration,
            'description': description
        })
        self.__handled_req(method='POST', uri=f"{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/mute",
                           chatmedia=True, data=data)

    def unmute_user(self, channel_url: str, user_id: str) -> None:
        self.__handled_req(method='DELETE', uri=f"{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/mute/{user_id}",
                           chatmedia=True)

    def set_channel_frozen_status(self, channel_url: str, is_frozen: bool) -> None:
        data = json.dumps({
            'freeze': is_frozen,
        })
        self.__handled_req(method='PUT',
                           uri=f"{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}/freeze",
                           chatmedia=True, data=data)

    def delete_channel(self, channel_url: str) -> None:
        self.__handled_req(method='DELETE',
                           uri=f"{SB_PROXY_CHATMEDIA}/v3/group_channels/{channel_url}",
                           chatmedia=True)



from typing import Dict, Optional, List
from enum import Enum


# AUTO GENERATED


class CustomType(str, Enum):
    group = 'group'
    direct = 'direct'


class MemberState(str, Enum):
    joined = 'joined'
    invited = 'invited'


class BannedUser:
    def __init__(self, in_data: dict):
        self.description: str = in_data.get('description')
        self.start_at: int = in_data.get('start_at')
        self.user: User = User(in_data.get('user'))
        self.end_at: int = in_data.get('end_at')


class BannedUsers:
    def __init__(self, in_data: dict):
        self.banned_list: List[BannedUser] = [BannedUser(n) for n in in_data.get('banned_list')]
        self.next: str = in_data.get('next')


class Channel:
    def __init__(self, in_data: dict):
        self.invited_at: Optional[int] = in_data.get('invited_at')
        self.custom_type: CustomType = CustomType[in_data.get('custom_type')]
        self.read_receipt: Dict[str] = in_data.get('read_receipt')
        self.member_state: MemberState = MemberState[in_data.get('member_state')]
        self.freeze: Optional[bool] = in_data.get('freeze')
        self.created_by: Optional[User] = User(in_data['created_by']) if in_data.get('created_by') is not None else None
        self.is_hidden: Optional[bool] = in_data.get('is_hidden')
        self.is_push_enabled: Optional[bool] = in_data.get('is_push_enabled')
        self.joined_ts: Optional[int] = in_data.get('joined_ts')
        self.is_created: Optional[bool] = in_data.get('is_created')
        self.member_count: Optional[int] = in_data.get('member_count')
        self.last_message: Optional[Message] = Message(in_data['last_message']) if in_data.get('last_message') is not None else None
        self.user_last_read: Optional[int] = in_data.get('user_last_read')
        self.unread_mention_count: Optional[int] = in_data.get('unread_mention_count')
        self.channel_url: str = in_data.get('channel_url')
        self.operators: Optional[List] = in_data.get('operators')
        self.channel: _Channel = _Channel(in_data.get('channel'))
        self.unread_message_count: Optional[int] = in_data.get('unread_message_count')
        self.cover_url: Optional[str] = in_data.get('cover_url')
        self.members: List[User] = [User(n) for n in in_data.get('members')]
        self.is_public: Optional[bool] = in_data.get('is_public')
        self.joined_member_count: Optional[int] = in_data.get('joined_member_count')
        self.is_super: Optional[bool] = in_data.get('is_super')
        self.name: Optional[str] = in_data.get('name')
        self.created_at: Optional[int] = in_data.get('created_at')
        self.max_length_message: Optional[int] = in_data.get('max_length_message')
        self.inviter: Optional[User] = User(in_data.get('inviter'))
        self.count_preference: Optional[str] = in_data.get('count_preference')


class Members:
    def __init__(self, in_data: dict):
        self.members: List[User] = [User(n) for n in in_data.get('members')]
        self.next: str = in_data.get('next')


class Message:
    def __init__(self, in_data: dict):
        self.mentioned_users: List[User] = [User(n) for n in in_data.get('mentioned_users')]
        self.updated_at: Optional[int] = in_data.get('updated_at')
        self.is_op_msg: Optional[bool] = in_data.get('is_op_msg')
        self.is_removed: Optional[bool] = in_data.get('is_removed')
        self.user: Optional[User] = User(in_data['user']) if in_data.get('user') is not None else None
        self.message: str = in_data.get('message')
        self.data: str = in_data.get('data')
        self.type: Optional[str] = in_data.get('type')
        self.created_at: Optional[int] = in_data.get('created_at')
        self.req_id: Optional[str] = in_data.get('req_id')
        self.mention_type: Optional[str] = in_data.get('mention_type')
        self.channel_url: str = in_data.get('channel_url')
        self.message_id: Optional[int] = in_data.get('message_id')


class User:
    def __init__(self, in_data: dict):
        self.is_blocking_me: Optional[bool] = in_data.get('is_blocking_me')
        self.user_id: Optional[str] = in_data.get('user_id')
        self.is_muted: Optional[bool] = in_data.get('is_muted')
        self.friend_name: Optional[str] = in_data.get('friend_name')
        self.joined_ts: Optional[int] = in_data.get('joined_ts')
        self.is_active: Optional[bool] = in_data.get('is_active')
        self.read_ts: Optional[int] = in_data.get('read_ts')
        self.is_blocked_by_me: Optional[bool] = in_data.get('is_blocked_by_me')
        self.state: Optional[str] = in_data.get('state')
        self.role: Optional[str] = in_data.get('role')
        self.is_online: Optional[bool] = in_data.get('is_online')
        self.require_auth_for_profile_image: Optional[bool] = in_data.get('require_auth_for_profile_image')
        self.nickname: Optional[str] = in_data.get('nickname')


class _Channel:
    def __init__(self, in_data: dict):
        self.name: str = in_data.get('name')
        self.member_count: int = in_data.get('member_count')
        self.custom_type: CustomType = CustomType[in_data.get('custom_type')]
        self.channel_url: str = in_data.get('channel_url')
        self.created_at: int = in_data.get('created_at')
        self.max_length_message: int = in_data.get('max_length_message')
        self.data: str = in_data.get('data')



import json
from types import SimpleNamespace
from enum import Enum


class FrameType(Enum):
    MESG = 'MESG'
    SYEV = 'SYEV'
    DELM = 'DELM'
    TPEN = 'TPEN'
    TPST = 'TPST'
    READ = 'READ'
    USEV = 'USEV'
    MACK = 'MACK'
    BRDM = 'BRDM'
    LOGI = 'LOGI'
    EROR = 'EROR'
    MEDI = 'MEDI'
    MRCT = 'MRCT'


class FrameModel(SimpleNamespace):
    def __repr__(self):
        keys = sorted(self.__dict__)
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "({})".format(", ".join(items))


def convert_to_framemodel(d):
    return json.loads(d, object_hook=lambda d: FrameModel(**d))


def get_frame_data(data):
    data_r = data[4:]
    data_j = convert_to_framemodel(data_r)
    type_f = FrameType[data[:4]]

    if type_f == FrameType.MESG:
        data_j.data = convert_to_framemodel(data_j.data)

    data_j.type_f = type_f

    return data_j



from typing import Callable, Optional
import websocket
from ._utils.rate_limiter import RateLimiter
import time
from ._utils.frame_model import get_frame_data, FrameType
from _thread import start_new_thread
from ._utils.ws_utils import get_ws_url, chat_printer, configure_loggers, pair_channel_and_names
from ._utils.consts import *


class WebSockClient:
    def __init__(self, access_token, user_id, enable_trace=False, print_chat=True, log_websocket_frames=False,
                 other_logging=True):
        self.user_id = user_id

        self.channelid_sub_pairs = {}
        self.RateLimiter = RateLimiter

        self.logger = configure_loggers()
        self.logger.disabled = not other_logging
        websocket.enableTrace(enable_trace)

        self.ws = self._get_ws_app(get_ws_url(self.user_id, access_token))

        self.req_id = int(time.time() * 1000)
        self.own_name = None
        self.print_chat = print_chat
        self.log_websocket_frames = log_websocket_frames
        self.last_err = None
        self.is_logi_err = False
        self.__session_key = None

        self.get_current_channels: Optional[Callable] = None
        self.current_channels = None

        self.after_message_hooks = []
        self.parralel_hooks = []

    def session_key_getter(self):
        return self.__session_key

    def _get_ws_app(self, ws_url):
        ws = websocket.WebSocketApp(ws_url,
                                    on_message=self.on_message,
                                    on_error=self.on_error,
                                    on_close=self.on_close,
                                    header={'User-Agent': USER_AGENT, 'Accept-Encoding': 'gzip'}
                                    )
        return ws

    def update_ws_app_urls_access_token(self, access_token):
        self.ws.url = get_ws_url(self.user_id, access_token)

    def on_message(self, _, message):
        resp = get_frame_data(message)
        if self.print_chat and resp.type_f == FrameType.MESG:
            chat_printer(resp, self.channelid_sub_pairs)
        if self.log_websocket_frames:
            self.logger.info(message)
        if resp.type_f == FrameType.LOGI:
            self.logger.info(message)
            self._logi(resp)

        for func in self.parralel_hooks:
            start_new_thread(func, (resp,))

        start_new_thread(self._response_loop, (resp,))

    def _logi(self, resp):
        try:
            logi_err = resp.error
        except AttributeError:
            logi_err = None
        if logi_err is None:
            self.is_logi_err = False
            self.last_err = None
            self.__session_key = resp.key
            self.update_channelid_sub_pair()
            self.own_name = resp.nickname
        else:
            self.logger.error(resp.message)
            self.is_logi_err = True

    def update_channelid_sub_pair(self):
        self.current_channels = self.get_current_channels(limit=100, order='latest_last_message', show_member=True,
                                                          show_read_receipt=True, show_empty=True,
                                                          member_state_filter='joined_only', super_mode='all',
                                                          public_mode='all', unread_filter='all',
                                                          hidden_mode='unhidden_only', show_frozen=True,
                                                          # custom_types='direct,group',
                                                          )
        self.channelid_sub_pairs = pair_channel_and_names(channels=self.current_channels,
                                                          own_user_id=self.user_id)

    def add_channelid_sub_pair(self, channel):
        self.current_channels.append(channel)
        self.channelid_sub_pairs = pair_channel_and_names(channels=self.current_channels,
                                                          own_user_id=self.user_id)

    def _response_loop(self, resp):
        for func in self.after_message_hooks:
            if func(resp):
                break

    def ws_send_message(self, text, channel_url):
        if self.RateLimiter.is_enabled and self.RateLimiter.check():
            return
        payload = MESG_regular.format(channel_url=channel_url, text=text, req_id=self.req_id)
        self.ws.send(payload)
        self.req_id += 1

    def ws_send_snoomoji(self, snoomoji, channel_url):
        if self.RateLimiter.is_enabled and self.RateLimiter.check():
            return
        payload = MESG_snoo.format(channel_url=channel_url, snoomoji=snoomoji, req_id=self.req_id)
        self.ws.send(payload)
        self.req_id += 1

    def ws_send_gif(self, gif_url, channel_url, height, width):
        payload = MESG_gif.format(gif_url=gif_url, channel_url=channel_url, height=height, width=width)
        self.ws.send(payload)

    def ws_send_img(self, img_url, channel_url, height, width, mimetype):
        payload = MESG_img.format(img_url=img_url, channel_url=channel_url, height=height,
                                  width=width, mimetype=mimetype)
        self.ws.send(payload)

    def ws_send_typing_indicator(self, channel_url):
        payload = TPST.format(channel_url=channel_url, time=int(time.time() * 1000))
        self.ws.send(payload)

    def ws_stop_typing_indicator(self, channel_url):
        payload = TPEN.format(channel_url=channel_url, time=int(time.time() * 1000))
        self.ws.send(payload)

    def on_error(self, _, error):
        self.logger.error(error)
        self.last_err = error

    def on_close(self, *_):
        self.logger.warning("### websocket closed ###")



from .ws_client import WebSockClient
import pickle
from .reddit_auth import _RedditAuthBase
from websocket import WebSocketConnectionClosedException
from ._api.tools import Tools
from ._api.models import Channel
from ._api.iconkeys import Snoo
from typing import Dict, List, Optional, Callable
from ._utils.frame_model import FrameType, FrameModel
from ._events import Events
from ._utils._exceptions import HookException


_hook = Callable[[FrameModel], Optional[bool]]


class ChatBot(Tools):
    def __init__(self, authentication: _RedditAuthBase, store_session: bool = True, log_error_frames=True,
                 **kwargs):
        self.__r_authentication = authentication
        self._store_session = store_session
        if store_session:
            sb_access_token, user_id = self._load_session(authentication._get_repr_pkl())
        else:
            reddit_authentication = self.__r_authentication.authenticate()
            sb_access_token, user_id = reddit_authentication['sb_access_token'], reddit_authentication['user_id']

        self.__WebSocketClient = WebSockClient(access_token=sb_access_token, user_id=user_id, **kwargs)
        super().__init__(self.__r_authentication, self.__WebSocketClient.session_key_getter, self._is_running)
        self.__WebSocketClient.get_current_channels = self.get_channels

        self.event = Events(self.__WebSocketClient)

        if log_error_frames:
            self.event.on_any(func=lambda resp: self.__WebSocketClient.logger.error(resp), frame_type=FrameType.EROR,
                              run_parallel=True)
        self.__is_running = False

    def _is_running(self) -> bool:
        return self.__is_running

    def get_own_name(self) -> str:
        return self.__WebSocketClient.own_name

    def get_own_userid(self) -> str:
        return self.__WebSocketClient.user_id

    def get_chatroom_name_id_pairs(self) -> Dict[str, str]:
        return self.__WebSocketClient.channelid_sub_pairs

    def get_channelurl_by_name(self, channel_name: str) -> str:
        return next(key for key, val in self.__WebSocketClient.channelid_sub_pairs.items() if val == channel_name)

    def set_respond_hook(self, input_: str,
                         response: str,
                         limited_to_users: List[str] = None,
                         lower_the_input: bool = False,
                         exclude_itself: bool = True,
                         must_be_equal: bool = True,
                         limited_to_channels: List[str] = None
                         ) -> None:
        if limited_to_channels is None:
            limited_to_channels = []
        if limited_to_users is None:
            limited_to_users = []
        try:
            response.format(nickname="")
        except KeyError:
            raise HookException("You need to set a {nickname} key in welcome message!")

        def hook(resp: FrameModel) -> Optional[bool]:
            sent_message = resp.message.lower() if lower_the_input else resp.message
            if (resp.user.name in limited_to_users or not bool(limited_to_users)) \
                    and (exclude_itself and resp.user.name != self.__WebSocketClient.own_name) \
                    and ((must_be_equal and sent_message == input_) or (not must_be_equal and input_ in sent_message)) \
                    and (self.__WebSocketClient.channelid_sub_pairs.get(
                resp.channel_url) in limited_to_channels or not bool(limited_to_channels)):
                response_prepped = response.format(nickname=resp.user.name)
                self.__WebSocketClient.ws_send_message(response_prepped, resp.channel_url)
                return True

        self.event.on_message(func=hook)

    def set_welcome_message(self, message: str, limited_to_channels: List[str] = None) -> None:
        if limited_to_channels is None:
            limited_to_channels = []
        try:
            message.format(nickname="", inviter="")
        except KeyError:
            raise HookException("Keys should be {nickname} and {inviter}")

        def hook(resp: FrameModel) -> Optional[bool]:
            if self.__WebSocketClient.channelid_sub_pairs.get(resp.channel_url) in limited_to_channels or \
                    not bool(limited_to_channels):
                response_prepped = message.format(nickname=resp.data.users[0].nickname,
                                                  inviter=resp.data.users[0].inviter.nickname)
                self.__WebSocketClient.ws_send_message(response_prepped, resp.channel_url)
                return True

        self.event.on_user_joined(hook)

    def set_farewell_message(self, message: str, limited_to_channels: List[str] = None) -> None:
        if limited_to_channels is None:
            limited_to_channels = []
        try:
            message.format(nickname="")
        except KeyError:
            raise HookException("Key should be {nickname}")

        def hook(resp: FrameModel) -> Optional[bool]:
            if self.__WebSocketClient.channelid_sub_pairs.get(resp.channel_url) in limited_to_channels or \
                    not bool(limited_to_channels):
                response_prepped = message.format(nickname=resp.data.nickname)
                self.__WebSocketClient.ws_send_message(response_prepped, resp.channel_url)
                return True

        self.event.on_user_left(hook)

    def remove_event_callback(self, func: _hook) -> None:
        self.__WebSocketClient.after_message_hooks.remove(func)

    def send_message(self, text: str, channel_url: str) -> None:
        self.__WebSocketClient.ws_send_message(text, channel_url)

    def send_snoomoji(self, snoomoji: Snoo, channel_url: str) -> None:
        self.__WebSocketClient.ws_send_snoomoji(snoomoji.value, channel_url)

    def send_gif(self, gif_url: str, channel_url: str, height: int = 200, width: int = 200) -> None:
        self.__WebSocketClient.ws_send_gif(gif_url, channel_url, height, width)

    def send_img(self, img_url: str, channel_url: str, height: int = 200, width: int = 200,
                 mimetype: str = "JPEG") -> None:
        self.__WebSocketClient.ws_send_img(img_url, channel_url, height, width, mimetype)

    def send_typing_indicator(self, channel_url: str) -> None:
        self.__WebSocketClient.ws_send_typing_indicator(channel_url)

    def stop_typing_indicator(self, channel_url: str) -> None:
        self.__WebSocketClient.ws_stop_typing_indicator(channel_url)

    def get_current_channels(self) -> List[Channel]:
        return self.__WebSocketClient.current_channels

    def run_4ever(self, auto_reconnect: bool = True, max_retries: int = 500, disable_ssl_verification: bool = False,
                  **kwargs) -> None:
        if disable_ssl_verification:
            import ssl
            sslopt = {"cert_reqs": ssl.CERT_NONE}
        else:
            sslopt = None

        self.__is_running = True
        for _ in range(max_retries):
            self.__WebSocketClient.ws.run_forever(ping_interval=15, ping_timeout=5,
                                                  skip_utf8_validation=True,
                                                  sslopt=sslopt,
                                                  **kwargs,
                                                  ping_payload="{active:1}"
                                                  )
            if self.__WebSocketClient.is_logi_err and self.__r_authentication.is_reauthable:
                self.__WebSocketClient.logger.info("Re-Authenticating...")
                if self._store_session:
                    sb_access_token, _ = self._load_session(self.__r_authentication._get_repr_pkl(), force_reauth=True)
                else:
                    sb_access_token = self.__r_authentication.authenticate()['sb_access_token']
                self.__WebSocketClient.update_ws_app_urls_access_token(sb_access_token)
            elif not (
                    auto_reconnect and isinstance(self.__WebSocketClient.last_err, WebSocketConnectionClosedException)):
                break
            self.__WebSocketClient.logger.info("Auto Re-Connecting...")

    def close(self) -> None:
        self.__WebSocketClient.ws.close()

    def leave_chat(self, channel_url: str) -> None:
        super().leave_chat(channel_url)
        self.__WebSocketClient.update_channelid_sub_pair()

    def delete_channel(self, channel_url: str) -> None:
        super().delete_channel(channel_url)
        self.__WebSocketClient.update_channelid_sub_pair()

    def get_chat_invites(self) -> List[Channel]:
        return self.get_channels(member_state_filter="invited_only")

    def create_channel(self, nicknames: List[str], group_name: str) -> Channel:
        nicknames.append(self.__WebSocketClient.own_name)
        channel = super().create_channel(nicknames, group_name)
        self.__WebSocketClient.add_channelid_sub_pair(channel)
        return channel

    def create_direct_channel(self, nickname: str) -> Channel:
        channel = self.create_channel(nicknames=[self.__WebSocketClient.own_name, nickname], group_name="")
        self.__WebSocketClient.add_channelid_sub_pair(channel)
        return channel

    def accept_chat_invite(self, channel_url: str) -> Channel:
        group = super().accept_chat_invite(channel_url)
        self.__WebSocketClient.add_channelid_sub_pair(group)
        return group

    def rename_channel(self, name: str, channel_url: str) -> Channel:
        channel = super().rename_channel(name, channel_url)
        self.__WebSocketClient.update_channelid_sub_pair()
        return channel

    def enable_rate_limiter(self, max_calls: float, period: float) -> None:
        self.__WebSocketClient.RateLimiter.is_enabled = True
        self.__WebSocketClient.RateLimiter.max_calls = max_calls
        self.__WebSocketClient.RateLimiter.period = period

    def disable_rate_limiter(self) -> None:
        self.__WebSocketClient.RateLimiter.is_enabled = False

    def _load_session(self, pkl_name, force_reauth=False):
        def get_store_file_handle(pkl_name_, mode_):
            try:
                return open(f"{pkl_name_}-stored.pkl", mode_)
            except FileNotFoundError:
                return None

        session_store_f = None if force_reauth else get_store_file_handle(pkl_name, 'rb')

        if session_store_f is None or force_reauth:
            session_store_f = get_store_file_handle(pkl_name, 'wb+')
            self.__r_authentication.authenticate()
            pickle.dump(self.__r_authentication, session_store_f)
        else:
            try:
                self.__r_authentication = pickle.load(session_store_f)
            except EOFError:
                return self._load_session(pkl_name, force_reauth=True)
        session_store_f.close()

        return self.__r_authentication.sb_access_token, self.__r_authentication.user_id



from setuptools import setup

setup(
    name='Reddit_ChatBot_Python',
    version='1.4.3',
    project_urls={"Source": "https://github.com/scrubjay55/Reddit_ChatBot_Python.git"},
    url="https://github.com/scrubjay55/Reddit_ChatBot_Python.git",
    author='scrubjay55',
    license="APLv2",
    packages=['Reddit_ChatBot_Python', 'Reddit_ChatBot_Python._utils', 'Reddit_ChatBot_Python._api'],
    description='an event-driven chatbot library for reddit chatrooms',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    python_requires='>=3.6',
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    install_requires=[
        "requests>=2.26.0",
        "websocket_client>=1.2.3",
    ],
)



from urllib.parse import urlencode
from .consts import SB_User_Agent, SB_ai
from .._api.models import Channel, CustomType
from typing import List
import logging


def get_ws_url(user_id: str, access_token: str):
    socket_base = "wss://sendbirdproxyk8s.chat.redditmedia.com"
    ws_params = {
        "user_id": user_id,
        "access_token": access_token,
        "p": "Android",
        "pv": 30,
        "sv": "3.0.144",
        "ai": SB_ai,
        "SB-User-Agent": SB_User_Agent,
        "active": "1"
    }
    return f"{socket_base}/?{urlencode(ws_params)}"


def pair_channel_and_names(channels: List[Channel], own_user_id: str):
    channelid_sub_pairs = {}
    for channel in channels:
        chn_name = channel.name
        if channel.custom_type == CustomType.direct:
            for mmbr in channel.members:
                if mmbr.user_id != own_user_id:
                    chn_name = mmbr.nickname
                    break
        channelid_sub_pairs.update({channel.channel_url: chn_name})
    return channelid_sub_pairs


def configure_loggers():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    logger.propagate = False
    sh = logging.StreamHandler()
    sh.setLevel(logging.INFO)
    sh.setFormatter(logging.Formatter(fmt="%(asctime)s, %(levelname)s: %(message)s", datefmt="%H:%M"))
    ws_logger = logging.getLogger("websocket")
    ws_logger.propagate = False
    ws_logger.addHandler(sh)
    logger.addHandler(sh)
    return logger


def chat_printer(resp, channelid_sub_pairs):
    if resp.message == "":
        try:
            msg = resp.data.v1.snoomoji
        except AttributeError:
            msg = resp.message
    else:
        msg = resp.message
    print(f"{resp.user.name}@{channelid_sub_pairs.get(resp.channel_url)}: {msg}")



from typing import Callable, Optional
from ._utils.frame_model import FrameType, FrameModel

_hook = Callable[[FrameModel], Optional[bool]]


class Events:
    def __init__(self, ws_client):
        self.__WebSocketClient = ws_client
        self.__ready_executed = False

    def on_any(self, func: _hook = None, frame_type: FrameType = FrameType.MESG, run_parallel=False):
        def wrap(func):
            def hook(resp: FrameModel):
                if resp.type_f == frame_type:
                    return func(resp)
            if run_parallel:
                self.__WebSocketClient.parralel_hooks.append(hook)
            else:
                self.__WebSocketClient.after_message_hooks.append(hook)
        
        if func is None:
            return wrap
        
        return wrap(func)

    def on_reaction(self, func: _hook = None, run_parralel=False):
        def wrap(func):
            return self.on_any(func, FrameType.MRCT, run_parralel)

        if func is None:
            return wrap

        return wrap(func)

    def on_image(self, func: _hook = None, run_parralel=False):
        def wrap(func):
            return self.on_any(func, FrameType.MEDI, run_parralel)

        if func is None:
            return wrap

        return wrap(func)

    def on_broadcast(self, func: _hook = None, run_parralel=False):
        def wrap(func):
            return self.on_any(func, FrameType.BRDM, run_parralel)

        if func is None:
            return wrap

        return wrap(func)

    def on_message(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            return self.on_any(func, FrameType.MESG, run_parallel)

        if func is None:
            return wrap

        return wrap(func)

    def on_ready(self, func: _hook = None):
        def wrap(func):
            def hook(resp: FrameModel) -> Optional[bool]:
                try:
                    _ = resp.error
                    return
                except AttributeError:
                    pass
                if self.__ready_executed:
                    return
                else:
                    self.__ready_executed = True
                return func(resp)

            return self.on_any(hook, FrameType.LOGI, False)

        if func is None:
            return wrap

        return wrap(func)

    def on_user_read(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            return self.on_any(func, FrameType.READ, run_parallel)

        if func is None:
            return wrap

        return wrap(func)

    def on_invitation(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            def hook(resp: FrameModel) -> Optional[bool]:
                try:
                    _ = resp.data.inviter
                    if not any([invitee.nickname == self.__WebSocketClient.own_name for invitee in resp.data.invitees]):
                        return
                except AttributeError:
                    return
                return func(resp)

            return self.on_any(hook, FrameType.SYEV, run_parallel)

        if func is None:
            return wrap

        return wrap(func)

    def on_invitation_of_other(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            def hook(resp: FrameModel) -> Optional[bool]:
                try:
                    _ = resp.data.inviter
                    if resp.cat != 10020:
                        return
                except AttributeError:
                    return

                return func(resp)

            return self.on_any(hook, FrameType.SYEV, run_parallel)

        if func is None:
            return wrap

        return wrap(func)

    def on_message_deleted(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            return self.on_any(func, FrameType.DELM, run_parallel)

        if func is None:
            return wrap

        return wrap(func)

    def on_user_joined(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            def hook(resp: FrameModel) -> Optional[bool]:
                try:
                    _ = resp.data.users[0].nickname
                    _ = resp.data.users[0].inviter.nickname
                except (AttributeError, IndexError):
                    return
                return func(resp)

            return self.on_any(hook, FrameType.SYEV, run_parallel)

        if func is None:
            return wrap

        return wrap(func)

    def on_user_left(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            def hook(resp: FrameModel) -> Optional[bool]:
                try:
                    _ = resp.channel.disappearing_message
                    _ = resp.data.nickname
                except AttributeError:
                    return
                return func(resp)

            return self.on_any(hook, FrameType.SYEV, run_parallel)

        if func is None:
            return wrap

        return wrap(func)

    def on_user_typing(self, func: _hook = None, run_parallel=False):
        def wrap(func):
            def hook(resp: FrameModel) -> Optional[bool]:
                try:
                    _ = resp.data.nickname
                    if resp.cat != 10900:
                        return
                except AttributeError:
                    return

                return func(resp)

            return self.on_any(hook, FrameType.SYEV, run_parallel)

        if func is None:
            return wrap

        return wrap(func)
